@@ -37,24 +37,25 @@
 #include <stdio.h>
 
 #define SUFFIX ".tablet"
-#define FEATURE_GROUP "Features"
+#define FEATURES_GROUP "Features"
 #define DEVICE_GROUP "Device"
+#define BUTTONS_GROUP "Buttons"
 
 static WacomClass
-libwacom_model_string_to_enum(const char *model)
+libwacom_class_string_to_enum(const char *class)
 {
-	if (model == NULL || *model == '\0')
+	if (class == NULL || *class == '\0')
 		return WCLASS_UNKNOWN;
 
-	if (strcmp(model, "Intuos3") == 0)
+	if (strcmp(class, "Intuos3") == 0)
 		return WCLASS_INTUOS3;
-	if (strcmp(model, "Intuos4") == 0)
+	if (strcmp(class, "Intuos4") == 0)
 		return WCLASS_INTUOS4;
-	if (strcmp(model, "Cintiq") == 0)
+	if (strcmp(class, "Cintiq") == 0)
 		return WCLASS_CINTIQ;
-	if (strcmp(model, "Bamboo") == 0)
+	if (strcmp(class, "Bamboo") == 0)
 		return WCLASS_BAMBOO;
-	if (strcmp(model, "Graphire") == 0)
+	if (strcmp(class, "Graphire") == 0)
 		return WCLASS_GRAPHIRE;
 
 	return WCLASS_UNKNOWN;
@@ -77,6 +78,8 @@ type_from_str (const char *type)
 		return WSTYLUS_MARKER;
 	if (strcmp (type, "Stroke") == 0)
 		return WSTYLUS_STROKE;
+	if (strcmp (type, "Puck") == 0)
+		return WSTYLUS_PUCK;
 	return WSTYLUS_UNKNOWN;
 }
 
@@ -167,9 +170,11 @@ libwacom_parse_stylus_keyfile(WacomDeviceDatabase *db, const char *path)
 				stylus->num_buttons = -1;
 				g_clear_error (&error);
 			}
+			stylus->has_lens = g_key_file_get_boolean(keyfile, groups[i], "HasLens", NULL);
 		} else {
 			stylus->num_buttons = 0;
 			stylus->has_eraser = FALSE;
+			stylus->has_lens = FALSE;
 		}
 
 		type = g_key_file_get_string(keyfile, groups[i], "Type", NULL);
@@ -184,6 +189,83 @@ libwacom_parse_stylus_keyfile(WacomDeviceDatabase *db, const char *path)
 	g_strfreev (groups);
 }
 
+struct {
+	const char       *key;
+	WacomButtonFlags  flag;
+} options[] = {
+	{ "Left", WACOM_BUTTON_POSITION_LEFT },
+	{ "Right", WACOM_BUTTON_POSITION_RIGHT },
+	{ "Top", WACOM_BUTTON_POSITION_TOP },
+	{ "Bottom", WACOM_BUTTON_POSITION_BOTTOM },
+	{ "Ring", WACOM_BUTTON_RING_MODESWITCH },
+	{ "Ring2", WACOM_BUTTON_RING2_MODESWITCH },
+	{ "Touchstrip", WACOM_BUTTON_TOUCHSTRIP_MODESWITCH },
+	{ "Touchstrip2", WACOM_BUTTON_TOUCHSTRIP2_MODESWITCH },
+	{ "OLEDs", WACOM_BUTTON_OLED }
+};
+
+static void
+libwacom_parse_buttons_key(WacomDevice      *device,
+			   GKeyFile         *keyfile,
+			   const char       *key,
+			   WacomButtonFlags  flag)
+{
+	guint i;
+	char **vals;
+
+	vals = g_key_file_get_string_list (keyfile, BUTTONS_GROUP, key, NULL, NULL);
+	if (vals == NULL)
+		return;
+	for (i = 0; vals[i] != NULL; i++) {
+		char val;
+
+		val = *vals[i];
+		if (strlen (vals[i]) > 1 ||
+		    val < 'A' ||
+		    val > 'Z') {
+			g_warning ("Ignoring value '%s' in key '%s'", vals[i], key);
+			continue;
+		}
+		val -= 'A';
+		device->buttons[(int) val] |= flag;
+	}
+
+	g_strfreev (vals);
+}
+
+static int
+libwacom_parse_num_modes (WacomDevice      *device,
+			  GKeyFile         *keyfile,
+			  const char       *key,
+			  WacomButtonFlags  flag)
+{
+	int num;
+	guint i;
+
+	num = g_key_file_get_integer (keyfile, BUTTONS_GROUP, key, NULL);
+	if (num > 0)
+		return num;
+	for (i = 0; i < device->num_buttons; i++) {
+		if (device->buttons[i] & flag)
+			num++;
+	}
+	return num;
+}
+
+static void
+libwacom_parse_buttons(WacomDevice *device,
+		       GKeyFile    *keyfile)
+{
+	guint i;
+
+	for (i = 0; i < G_N_ELEMENTS (options); i++)
+		libwacom_parse_buttons_key(device, keyfile, options[i].key, options[i].flag);
+
+	device->ring_num_modes = libwacom_parse_num_modes(device, keyfile, "RingNumModes", WACOM_BUTTON_RING_MODESWITCH);
+	device->ring2_num_modes = libwacom_parse_num_modes(device, keyfile, "Ring2NumModes", WACOM_BUTTON_RING2_MODESWITCH);
+	device->strips_num_modes = libwacom_parse_num_modes(device, keyfile, "StripsNumModes", WACOM_BUTTON_TOUCHSTRIP_MODESWITCH);
+}
+
 static WacomDevice*
 libwacom_parse_tablet_keyfile(const char *path)
 {
@@ -217,17 +299,16 @@ libwacom_parse_tablet_keyfile(const char *path)
 			device = NULL;
 			goto out;
 		}
-		device->match = g_strdup_printf ("%s:0x%x:0x%x", bus_to_str (device->bus), device->vendor_id, device->product_id);
+		device->match = g_strdup_printf ("%s:%04x:%04x", bus_to_str (device->bus), device->vendor_id, device->product_id);
 		g_free (match);
 	}
 
-	device->vendor = g_key_file_get_string(keyfile, DEVICE_GROUP, "Vendor", NULL);
-	device->product = g_key_file_get_string(keyfile, DEVICE_GROUP, "Product", NULL);
+	device->name = g_key_file_get_string(keyfile, DEVICE_GROUP, "Name", NULL);
 	device->width = g_key_file_get_integer(keyfile, DEVICE_GROUP, "Width", NULL);
 	device->height = g_key_file_get_integer(keyfile, DEVICE_GROUP, "Height", NULL);
 
 	class = g_key_file_get_string(keyfile, DEVICE_GROUP, "Class", NULL);
-	device->cls = libwacom_model_string_to_enum(class);
+	device->cls = libwacom_class_string_to_enum(class);
 	g_free(class);
 
 	styli_list = g_key_file_get_string_list(keyfile, DEVICE_GROUP, "Styli", NULL, NULL);
@@ -249,31 +330,43 @@ libwacom_parse_tablet_keyfile(const char *path)
 	}
 
 	/* Features */
-	if (g_key_file_get_boolean(keyfile, FEATURE_GROUP, "Stylus", NULL))
+	if (g_key_file_get_boolean(keyfile, FEATURES_GROUP, "Stylus", NULL))
 		device->features |= FEATURE_STYLUS;
 
-	if (g_key_file_get_boolean(keyfile, FEATURE_GROUP, "Touch", NULL))
+	if (g_key_file_get_boolean(keyfile, FEATURES_GROUP, "Touch", NULL))
 		device->features |= FEATURE_TOUCH;
 
-	if (g_key_file_get_boolean(keyfile, FEATURE_GROUP, "Ring", NULL))
+	if (g_key_file_get_boolean(keyfile, FEATURES_GROUP, "Ring", NULL))
 		device->features |= FEATURE_RING;
 
-	if (g_key_file_get_boolean(keyfile, FEATURE_GROUP, "Ring2", NULL))
+	if (g_key_file_get_boolean(keyfile, FEATURES_GROUP, "Ring2", NULL))
 		device->features |= FEATURE_RING2;
 
-	if (g_key_file_get_boolean(keyfile, FEATURE_GROUP, "VStrip", NULL))
-		device->features |= FEATURE_VSTRIP;
-
-	if (g_key_file_get_boolean(keyfile, FEATURE_GROUP, "HStrip", NULL))
-		device->features |= FEATURE_HSTRIP;
-
-	if (g_key_file_get_boolean(keyfile, FEATURE_GROUP, "BuiltIn", NULL))
+	if (g_key_file_get_boolean(keyfile, FEATURES_GROUP, "BuiltIn", NULL))
 		device->features |= FEATURE_BUILTIN;
 
-	if (g_key_file_get_boolean(keyfile, FEATURE_GROUP, "Reversible", NULL))
+	if (g_key_file_get_boolean(keyfile, FEATURES_GROUP, "Reversible", NULL))
 		device->features |= FEATURE_REVERSIBLE;
 
-	device->num_buttons = g_key_file_get_integer(keyfile, FEATURE_GROUP, "Buttons", NULL);
+	if (device->features & FEATURE_BUILTIN &&
+	    device->features & FEATURE_REVERSIBLE)
+		g_warning ("Tablet '%s' is both reversible and builtin. This is impossible", device->match);
+
+	if (!(device->features & FEATURE_RING) &&
+	    (device->features & FEATURE_RING2))
+		g_warning ("Table '%s' has Ring2 but no Ring. This is impossible", device->match);
+
+	device->num_strips = g_key_file_get_integer(keyfile, FEATURES_GROUP, "NumStrips", NULL);
+	device->num_buttons = g_key_file_get_integer(keyfile, FEATURES_GROUP, "Buttons", &error);
+	if (device->num_buttons == 0 &&
+	    g_error_matches (error, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_KEY_NOT_FOUND)) {
+		g_warning ("Tablet '%s' has no buttons defined, do something!", device->match);
+		g_clear_error (&error);
+	}
+	if (device->num_buttons > 0) {
+		device->buttons = g_new0 (WacomButtonFlags, device->num_buttons);
+		libwacom_parse_buttons(device, keyfile);
+	}
 
 out:
 	if (keyfile)
