@@ -40,10 +40,11 @@ libwacom_get_device(WacomDeviceDatabase *db, const char *match)
 	return (WacomDevice *) g_hash_table_lookup (db->device_ht, match);
 }
 
-static int
+static gboolean
 get_device_info (const char   *path,
 		 int          *vendor_id,
 		 int          *product_id,
+		 char        **name,
 		 WacomBusType *bus,
 		 IsBuiltin    *builtin,
 		 WacomError   *error)
@@ -57,6 +58,7 @@ get_device_info (const char   *path,
 
 	retval = FALSE;
 	*builtin = IS_BUILTIN_UNSET;
+	*name = NULL;
 	client = g_udev_client_new (subsystems);
 	device = g_udev_client_query_by_device_file (client, path);
 	if (device == NULL) {
@@ -70,6 +72,11 @@ get_device_info (const char   *path,
 	}
 
 	bus_str = g_udev_device_get_property (device, "ID_BUS");
+	/* Serial devices are weird */
+	if (bus_str == NULL) {
+		if (g_strcmp0 (g_udev_device_get_subsystem (device), "tty") == 0)
+			bus_str = "serial";
+	}
 	/* Poke the parent device for Bluetooth models */
 	if (bus_str == NULL) {
 		GUdevDevice *parent;
@@ -81,7 +88,6 @@ get_device_info (const char   *path,
 		bus_str = "bluetooth";
 	}
 
-
 	/* Is the device builtin? */
 	devname = g_udev_device_get_name (device);
 	if (devname != NULL) {
@@ -99,6 +105,16 @@ get_device_info (const char   *path,
 		g_free (sysfs_path);
 	}
 
+	*name = g_strdup (g_udev_device_get_sysfs_attr (device, "name"));
+	/* Try getting the name from the parent if that fails */
+	if (*name == NULL) {
+		GUdevDevice *parent;
+
+		parent = g_udev_device_get_parent (device);
+		*name = g_strdup (g_udev_device_get_sysfs_attr (parent, "name"));
+		g_object_unref (parent);
+	}
+
 	*bus = bus_from_str (bus_str);
 	if (*bus == WBUSTYPE_USB) {
 		const char *vendor_str, *product_str;
@@ -118,14 +134,16 @@ get_device_info (const char   *path,
 		 * vendor 0x56a
 		 * product 0x81 */
 		product_str = g_udev_device_get_property (device, "PRODUCT");
+		g_assert (product_str);
 		if (sscanf(product_str, "%d/%x/%x/%d", &garbage, vendor_id, product_id, &garbage) != 4) {
 			libwacom_error_set(error, WERROR_UNKNOWN_MODEL, "Unimplemented serial bus");
 			goto bail;
 		}
 	} else if (*bus == WBUSTYPE_SERIAL) {
-		/* FIXME implement */
-		libwacom_error_set(error, WERROR_UNKNOWN_MODEL, "Unimplemented serial bus");
-		goto bail;
+		/* FIXME This matches the declaration in serial-wacf004.tablet
+		 * Might not be good enough though */
+		vendor_id = 0;
+		product_id = 0;
 	} else {
 		libwacom_error_set(error, WERROR_UNKNOWN_MODEL, "Unsupported bus '%s'", bus_str);
 		goto bail;
@@ -135,8 +153,13 @@ get_device_info (const char   *path,
 	    vendor_id != 0 &&
 	    product_id != 0)
 		retval = TRUE;
+	/* The serial bus uses 0:0 as the vid/pid */
+	if (*bus == WBUSTYPE_SERIAL)
+		retval = TRUE;
 
 bail:
+	if (retval == FALSE)
+		g_free (*name);
 	if (device != NULL)
 		g_object_unref (device);
 	if (client != NULL)
@@ -150,8 +173,7 @@ libwacom_copy(const WacomDevice *device)
 	WacomDevice *d;
 
 	d = g_new0 (WacomDevice, 1);
-	d->vendor = g_strdup (device->vendor);
-	d->product = g_strdup (device->product);
+	d->name = g_strdup (device->name);
 	d->width = device->width;
 	d->height = device->height;
 	d->match = g_strdup (device->match);
@@ -159,11 +181,15 @@ libwacom_copy(const WacomDevice *device)
 	d->product_id = device->product_id;
 	d->cls = device->cls;
 	d->bus = device->bus;
-	d->num_buttons = device->num_buttons;
-	d->supported_styli = g_memdup (device->supported_styli, sizeof(int) * device->num_styli);
-	d->num_styli = device->num_styli;
+	d->num_strips = device->num_strips;
 	d->features = device->features;
-
+	d->strips_num_modes = device->strips_num_modes;
+	d->ring_num_modes = device->ring_num_modes;
+	d->ring2_num_modes = device->ring2_num_modes;
+	d->num_styli = device->num_styli;
+	d->supported_styli = g_memdup (device->supported_styli, sizeof(int) * device->num_styli);
+	d->num_buttons = device->num_buttons;
+	d->buttons = g_memdup (device->buttons, sizeof(WacomButtonFlags) * device->num_buttons);
 	return d;
 }
 
@@ -178,7 +204,7 @@ libwacom_new (WacomDeviceDatabase *db, int vendor_id, int product_id, WacomBusTy
         return NULL;
     }
 
-    match = g_strdup_printf ("%s:0x%x:0x%x", bus_to_str (bus), vendor_id, product_id);
+    match = g_strdup_printf ("%s:%04x:%04x", bus_to_str (bus), vendor_id, product_id);
     device = libwacom_get_device(db, match);
     g_free (match);
 
@@ -191,26 +217,50 @@ libwacom_new_from_path(WacomDeviceDatabase *db, const char *path, int fallback,
     int vendor_id, product_id;
     WacomBusType bus;
     const WacomDevice *device;
+    WacomDevice *ret;
     IsBuiltin builtin;
+    char *name;
+
+    if (!db) {
+        libwacom_error_set(error, WERROR_INVALID_DB, "db is NULL");
+        return NULL;
+    }
 
     if (!path) {
         libwacom_error_set(error, WERROR_INVALID_PATH, "path is NULL");
         return NULL;
     }
 
-    if (!get_device_info (path, &vendor_id, &product_id, &bus, &builtin, error))
+    if (!get_device_info (path, &vendor_id, &product_id, &name, &bus, &builtin, error))
         return NULL;
 
     device = libwacom_new (db, vendor_id, product_id, bus, error);
+    if (device != NULL)
+	    ret = libwacom_copy(device);
+    else if (!fallback)
+	    goto bail;
 
-    if (device == NULL && fallback)
+    if (device == NULL && fallback) {
 	    device = libwacom_get_device(db, "generic");
-
-    if (device) {
-	    WacomDevice *ret;
+	    if (device == NULL)
+		    goto bail;
 
 	    ret = libwacom_copy(device);
 
+	    if (name != NULL) {
+		    g_free (ret->name);
+		    ret->name = name;
+	    }
+	    ret->vendor_id = vendor_id;
+	    ret->product_id = product_id;
+	    ret->bus = bus;
+	    g_free (ret->match);
+	    ret->match = g_strdup_printf ("%s:0x%x:0x%x", bus_to_str (bus), vendor_id, product_id);
+    } else {
+	    g_free (name);
+    }
+
+    if (device) {
 	    if (builtin == IS_BUILTIN_TRUE)
 	        ret->features |= FEATURE_BUILTIN;
 	    else if (builtin == IS_BUILTIN_FALSE)
@@ -219,6 +269,8 @@ libwacom_new_from_path(WacomDeviceDatabase *db, const char *path, int fallback,
 	    return ret;
     }
 
+bail:
+    g_free (name);
     libwacom_error_set(error, WERROR_UNKNOWN_MODEL, NULL);
     return NULL;
 }
@@ -228,6 +280,11 @@ libwacom_new_from_usbid(WacomDeviceDatabase *db, int vendor_id, int product_id,
 {
     const WacomDevice *device;
 
+    if (!db) {
+        libwacom_error_set(error, WERROR_INVALID_DB, "db is NULL");
+        return NULL;
+    }
+
     device = libwacom_new(db, vendor_id, product_id, WBUSTYPE_USB, error);
 
     if (device)
@@ -253,7 +310,7 @@ libwacom_new_from_name(WacomDeviceDatabase *db, const char *name, WacomError *er
     for (l = keys; l; l = l->next) {
         WacomDevice *d = l->data;
 
-        if (g_strcmp0 (d->product, name) == 0) {
+        if (g_strcmp0 (d->name, name) == 0) {
             device = d;
             break;
 	}
@@ -270,27 +327,22 @@ libwacom_new_from_name(WacomDeviceDatabase *db, const char *name, WacomError *er
 void
 libwacom_destroy(WacomDevice *device)
 {
-	g_free (device->vendor);
-	g_free (device->product);
+	g_free (device->name);
 
 	g_free (device->match);
 	g_free (device->supported_styli);
+	g_free (device->buttons);
 	g_free (device);
 }
 
-const char* libwacom_get_vendor(WacomDevice *device)
-{
-    return device->vendor;
-}
-
 int libwacom_get_vendor_id(WacomDevice *device)
 {
     return device->vendor_id;
 }
 
-const char* libwacom_get_product(WacomDevice *device)
+const char* libwacom_get_name(WacomDevice *device)
 {
-    return device->product;
+    return device->name;
 }
 
 int libwacom_get_product_id(WacomDevice *device)
@@ -352,14 +404,24 @@ int libwacom_has_ring2(WacomDevice *device)
     return !!(device->features & FEATURE_RING2);
 }
 
-int libwacom_has_vstrip(WacomDevice *device)
+int libwacom_get_ring_num_modes(WacomDevice *device)
 {
-    return !!(device->features & FEATURE_VSTRIP);
+    return device->ring_num_modes;
 }
 
-int libwacom_has_hstrip(WacomDevice *device)
+int libwacom_get_ring2_num_modes(WacomDevice *device)
 {
-    return !!(device->features & FEATURE_HSTRIP);
+    return device->ring2_num_modes;
+}
+
+int libwacom_get_num_strips(WacomDevice *device)
+{
+    return device->num_strips;
+}
+
+int libwacom_get_strips_num_modes(WacomDevice *device)
+{
+    return device->strips_num_modes;
 }
 
 int libwacom_is_builtin(WacomDevice *device)
@@ -377,6 +439,21 @@ WacomBusType libwacom_get_bustype(WacomDevice *device)
     return device->bus;
 }
 
+WacomButtonFlags
+libwacom_get_button_flag(WacomDevice *device,
+			 char         button)
+{
+	int index;
+
+	g_return_val_if_fail (device->num_buttons > 0, WACOM_BUTTON_NONE);
+	g_return_val_if_fail (button >= 'A', WACOM_BUTTON_NONE);
+	g_return_val_if_fail (button < 'A' + device->num_buttons, WACOM_BUTTON_NONE);
+
+	index = button - 'A';
+
+	return device->buttons[index];
+}
+
 const WacomStylus *libwacom_stylus_get_for_id (WacomDeviceDatabase *db, int id)
 {
 	return g_hash_table_lookup (db->stylus_ht, GINT_TO_POINTER(id));
@@ -411,6 +488,11 @@ int libwacom_stylus_is_eraser (const WacomStylus *stylus)
 	return stylus->is_eraser;
 }
 
+int libwacom_stylus_has_lens (const WacomStylus *stylus)
+{
+	return stylus->has_lens;
+}
+
 WacomStylusType libwacom_stylus_get_type (const WacomStylus *stylus)
 {
 	if (stylus->type == WSTYLUS_UNKNOWN) {
